// matrix.h:         Darstellung von Matrizen
// -----------------------------------------------------
#include <valarray>
#include <stdexcept>
using namespace std;

template<class T>
class Matrix
{
  private:
    size_t rows, cols;      // Anzahl Zeilen und Spalten.
    valarray<T> element;    // Numerischer Vektor
  public:
    Matrix(size_t n, size_t m, const T& a = T(0))
         : rows(n), cols(m), element(a, n*m){}
    Matrix(const T *v, size_t n, size_t m)
         : rows(n), cols(m), element(v, n*m){}
    Matrix(const valarray<T>& v, size_t n, size_t m)
         : rows(n), cols(m), element(v){}

    size_t getRows() const { return rows; }
    size_t getCols() const { return cols; }

    T& operator()(size_t i, size_t j)  // Element in i-ter Zeilen
    {                                  // j-ter Spalte.
        if( i < rows && j < cols)
          return element[i*cols + j];
        else
          throw out_of_range("Unzulaessiger Index!");
    }

    valarray<T> operator[](size_t i) const    // i. Zeilenvektor
    {
      if( i < rows)
          return element[slice(i*cols, cols, 1)];
      else
         throw out_of_range("Unzulaessiger Zeilenindex!");
    }

    // Zuweisung, Operatoren etc.

    Matrix subMatrix(size_t is, size_t js, size_t ie, size_t je)
    {
      if( is <= ie && ie < rows && js <= je && je < cols)
      {
        size_t srows = ie - is + 1,   scols = je - js + 1;
        valarray<size_t> len{ srows, scols},
                         dist{cols, 1};
        valarray<T> gsl(element[gslice(is*cols+js, len, dist)]);
        return Matrix(gsl, srows, scols);
      }
      else throw out_of_range("Unzulaessiger Index!");
    }

   Matrix minor(size_t i, size_t j)
   {
     if( i < rows && j < cols)
     {
       valarray<size_t> dist{cols-1, 1},    // Schrittweiten fr
                        len(2);             // neue Matrix.
       Matrix<T> adj(rows - 1, cols - 1);   // Teilmatrix

       if(i != 0 && j !=  0)                // links oberhalb [i,j]
       {
         len[0] = i; len[1] = j;
         Matrix<T> subMat = subMatrix(0,0,i-1,j-1);
         adj.element[gslice(0, len, dist)] = subMat.element;
       }
       if( i != 0 && j != cols-1)         // rechts oberhalb [i,j]
       {
         len[0]= i; len[1] = cols - j - 1;
         Matrix<T> subMat = subMatrix(0,j+1,i-1,cols-1);
         adj.element[gslice(j, len, dist)] = subMat.element;
       }
       if(i != rows-1 && j != 0)          // links unterhalb [i,j]
       {
         len[0]= rows - i - 1; len[1] = j;
         Matrix<T> subMat = subMatrix(i+1,0,rows-1,j-1);
         adj.element[gslice(i*(cols-1), len, dist)] = subMat.element;
       }
       if( i!=rows-1  && j != cols-1)     // rechts unterhalb [i,j]
       {
         len[0]= rows - i - 1; len[1] = cols - j - 1;
         Matrix<T> subMat= subMatrix(i+1,j+1,rows-1,cols-1);
         adj.element[gslice(i*(cols-1)+j,len,dist)] = subMat.element;
       }
     return adj;
     }
     else throw out_of_range("Unzulaessiger Index!");
   }

   Matrix transponierte()      // An der "Diagonalen" spiegeln.
   {
      // Indirekte Indizierung:
      valarray<size_t> indirect(rows*cols);
      for(size_t i = 0; i < rows; i++)
         for(size_t j = 0; j < cols; j++)
             indirect[j*rows + i] = i*cols + j;

      Matrix trans(element[indirect], cols, rows);
      return trans;
   }

   Matrix& upperTriangle(const Matrix& m)
   {
     if( rows != m.rows || cols != m.cols)
         throw length_error("Matrix nicht kompatibel.");
     if( rows != cols )
         throw length_error("Matrix nicht quadratisch");

     // Maske:
     valarray<bool> mask(false, cols*cols);
     for(size_t i = 0; i < rows; ++i)   // Obere Dreiecksmatrix
        for(size_t j = i+1; j < rows; ++j)
           mask[i*cols + j] = true;

     // Maskierte Indizierung:
     element[mask] = m.element;
     return *this;
   }
};
