// mySmartPtr.h
// Smart Pointer auf dynamische Objekte mit Referenzzhler.
// Das Objekt wird zerstrt, wenn der letzte Zeiger darauf 
// zerstrt wird.
// ----------------------------------------------------------
#ifndef MYSMARTPTR_H
#define MYSMARTPTR_H

#include <assert.h>

template <class T>
class SmartPtr
{
  private: 
    T *objPtr;           // Zeiger auf das dynamische Objekt.
    int* countPtr;       // Referenzzhler (Anzahl Besitzer
                         // = Anzahl Zeiger auf das Objekt).
  public:
    SmartPtr() : objPtr(0), countPtr(0) {}   // Leerer Smart Pointer
    template <class U>
    explicit SmartPtr(U *ptr) : objPtr((T*)ptr), countPtr(new int(1))
    {}

    // Kopierkonstruktor (ein neuer Besitzer)
    SmartPtr(const SmartPtr& p)
    : objPtr(p.objPtr), countPtr(p.countPtr)
    { if(countPtr) ++*countPtr; }       // Falls p nicht leer.

    // Kopierkonstruktor mit impliziter Konvertierung
    template <class U>
    SmartPtr(const SmartPtr<U>& p)
    : objPtr((T*)p.getObjPtr()), countPtr((int*)p.getCountPtr())
    { if(countPtr) ++*countPtr; }       // falls p nicht leer.

    // Zuweisung
    SmartPtr& operator=(const SmartPtr& p)
    {
       if( objPtr != p.objPtr)
       {
           reset(); 
           objPtr   = p.objPtr;
           countPtr = p.countPtr;
           if(countPtr) ++*countPtr;     // Falls p nicht leer.
       }
       return *this;
    }

    // Zuweisung mit impliziter Konvertierung
    template <class U>
    SmartPtr& operator=(const SmartPtr<U>& p)
    {
       if( objPtr != p.getObjPtr())
       {
           reset(); 
           objPtr   = (T*)p.getObjPtr();
           countPtr = (int*)p.getCountPtr();
           if(countPtr) ++*countPtr;     // Falls p nicht leer.
       }
       return *this;
    }

    // Destruktor 
    ~SmartPtr() { reset(); }

    // Objekt freigeben und als letzter Besitzer lschen:
    void reset();

    // Objektzugriffe:
    T* operator->() const
    {
       assert(objPtr);
       return objPtr;
    }
    T& operator*() const
    {
       assert(objPtr);
       return *objPtr;
    }

    T* getObjPtr() const  { return objPtr; }
    const int* getCountPtr() const  { return countPtr; }

    // explicit 
    operator bool() const { return objPtr != 0; }

    void swap( SmartPtr& p)
    { 
       T* op = objPtr;     objPtr = p.objPtr;     p.objPtr = op;
       int* cp = countPtr; countPtr = p.countPtr; p.countPtr = cp;
    }
};

template <class T>
inline void SmartPtr<T>::reset()     // Der letzte Besitzer lscht das Objekt
{                                    // und den Zaehler. 
    if( countPtr && --*countPtr == 0)
    {
        delete objPtr; delete countPtr;
        objPtr = 0; countPtr = 0;
    }
}

#endif    // MYSMARTPTR_H
